import{p as e,f as o}from"./flowDb-956e92f1-bf75726a.js";import{f as t,g as a}from"./styles-c10674c1-8255ce96.js";import{u as i}from"./index-a5a54ef8.js";import"./graph-bfdb8d0f.js";import"./layout-96a8a310.js";import"./index-3862675e-71e1acae.js";import"./clone-4ba2ff89.js";import"./edges-e0da2a9e-eb805b1b.js";import"./createText-2e5e7dd3-994fe037.js";import"./line-d7ccdbaa.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-8b49a465.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
